/******************************************************************************/
/**
@file "MPI/C/Primer 1 - empty/prazan_primer_mpi.c"
@brief Genericki primer MPI programa.
*/
/******************************************************************************/

#include <mpi.h>

/******************************************************************************/

/**
@def WORK_TAG
@brief Opis poruka koje se salju ostalim procesima.
*/
#define WORK_TAG 1

/**
@def DIE_TAG
@brief Opis poruke koja ubija ostale procese.
*/
#define DIE_TAG 2

/******************************************************************************/

/**
@details
Ukoliko smo master proces sa pristupom konzoli poziva se ova funkcija.
*/
static void master(void);

/**
@details
Ukoliko nismo master proces sa pristupom konzoli poziva se ova funkcija.
*/
static void slave(void);

/**
@return Vraca deo posla koji je sledeci na redu za izvrsavanje.
@details
Funkcija koja odredjuje koji deo posla je sledeci na redu za izvrsavanje.
*/
static unit_of_work_t get_next_work_item(void);

/**
@details
Ukoliko nismo master proces sa pristupom konzoli poziva se ova funkcija.
*/
static void process_results(unit_result_t result);

/**
@details
Funkcija koja vrsi neku obradu.
*/
static unit_result_t do_work(unit_of_work_t work);

/******************************************************************************/

/**
@brief Glavni program
@param argc Broj argumenata komandne linije.
@param argv Argumenti komandne linije.
@return Vraca @b 0 ako je sve uspesno izvrseno.
*/
int main(int argc, char** argv)
{
	int my_rank;
	
	/* Inicijalizujemo MPI */
	
	MPI_Init(&argc, &argv);
	
	/* Pronadjemo nas rank u MPI svetu */	
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

	if (my_rank == 0)
	{
		master();
	}
	else
	{
		slave();
	}
	
	/* Zavrsavamo MPI */	
	MPI_Finalize();

	return 0;
}

/******************************************************************************/

static void master(void)
{
	int ntasks, rank;
	unit_of_work_t work;
	unit_result_t result;
	MPI_Status status;
	
	/* Utvrdimo velicinu MPI sveta */
	MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
	/* Svakom slugi u MPI svetu posaljemo odredjeni deo posla (seed-ujemo ih) */
	for (rank = 1; rank < ntasks; ++rank)
	{
		/* Odredimo koji je sledeci deo posla */
		work = get_next_work_item();
	
		/* Posaljemo deo posla sledecem procesu */	
		MPI_Send(&work,         /* Bafer koji sadrzi poruku */
			 1,	                /* Broj podataka koje saljemo */
			 MPI_INT,           /* Podatak koji saljemo je ceo broj */
			 rank,              /* Rank procesa koji prima poruku */
			 WORK_TAG,          /* Oznaka poruke koju je odabrao korisnik */
			 MPI_COMM_WORLD);   /* MPI svet */
	}
	
	/*
	Vrtimo se u petlji i saljemo delove poslova ostalim procesima,
	sve dok ima posla
	*/
	work = get_next_work_item();
	while (work != NULL)
	{		
		/* Primamo rezultate od sluge */		
		MPI_Recv(&result,		/* Bafer u koji smestamo poruku */
			1,			/* Broj podataka koje primamo */
			MPI_DOUBLE,		/* Tip podataka koje primamo */
			MPI_ANY_SOURCE,		/* Svaki izvor poruka je dozvoljen */
			MPI_ANY_TAG,		/* Svi tipovi poruka su dozvoljeni */
			MPI_COMM_WORLD,		/* MPI svet*/
			&status);		/* Informacija o primljenoj poruci */
		
		/* Posaljemo slugi novi deo posla */
		MPI_Send(&work,			/* Bafer koji sadrzi poruku */
			 1,			/* Broj podataka koje saljemo */
			 MPI_INT,		/* Podatak koji saljemo je ceo broj */
			 status.MPI_SOURCE,	/* Saljemo onome od koga smo upravo primili poruku */
			 WORK_TAG,		/* Oznaka poruke koju je odabrao korisnik */
			 MPI_COMM_WORLD);   	/* MPI svet */	
		
		/* Odredimo koji je sledeci deo posla */		
		work = get_next_work_item();
	}
	/* Posto nema vise posla treba prikupiti rezultate od sluga */	
	for (rank = 1; rank < ntasks; ++rank)
	{
		MPI_Recv(&result, 1, MPI_DOUBLE, MPI_ANY_SOURCE,
			MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	}

	/* Obradimo prikupljene rezultate */
	process_results(result);

	/* Slanjem prazne poruke sa DIE_TAG opisom kazemo svim slugama da zavrse */
	for (rank = 1; rank < ntasks; ++rank)
	{
		MPI_Send(0, 0, MPI_INT, rank, DIE_TAG, MPI_COMM_WORLD);
	}
}	

/******************************************************************************/

static void slave(void)
{
	/* Bafer u kome se cuva opis dela posla koji treba odraditi */
	unit_of_work_t work;
	unit_result_t results;
	MPI_Status status;
	
	while (1)
	{	
		/* Primamo deo posla od mastera */		
		MPI_Recv(&work, 1, MPI_INT, 0, MPI_ANY_TAG,
			MPI_COMM_WORLD, &status);
		
		/* Proveravamo opsi dobijene poruke */		
		if (status.MPI_TAG == DIE_TAG)
			/* Ako je opis jednak DIE_TAG, ciao tutti! */
			return;
	
		/* Pozovemo rutinu koja radi posao */
		result = do_work(work);	
		/* Vratimo rezultate masteru */
		MPI_Send(&result, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
	}
}

/******************************************************************************/

static unit_of_work_t get_next_work_item(void)
{
/* Ubaciti neki kod */
}

/******************************************************************************/

static void process_results(unit_result_t result)
{
/* Ubaciti neki kod */
}

/******************************************************************************/

static unit_result_t do_work(unit_of_work_t work)
{
/* Ubaciti neki kod */
}

/******************************************************************************/
