/**
@file "quicksort_openmp.cpp"
@brief Paralelna implementacija algoritma Quicksort pomocu OpenMP.
		
	   Posto se radi o rekurzivnom algoritmu, u svakoj iteraciji se stvara novi (ugnezdeni) tim od 2 niti,
	   i svaka od tih niti dalje nastavlja da sortira svoj deo niza. Vodi se racuna da ukupan broj niti u
	   procesu ne premasi gornju granicu broja niti (koja bi trebala biti <= broju dostupnih procesora),
	   jer bi u suprotnom doslo do gubitka performansi zbog stvaranja ogromnog broja niti.
	   
	   Iako OpenMP standard dozvoljava ugnezdavanje niti, to trenutno ne podrzavaju svi kopmajeri. 
	   Recimo, Intel-ov kompajler podrzava, dok Microsoft C++ kompajler ne podrzava, pa to treba imati u vidu
	   prilikom merenja performansi.
	   
	   Za optimalne performanse, algoritam se mogao modifikovati tako da se predje na neki drugi algoritam
	   za sortiranje koji je optimalniji za manje nizove, kada velicina particije opadne ispod odredjene granice.
	   Dalje ubrzanje je moguce postici razvijajuci rekurziju preko steka u iterativni postupak.
	   
	   
*/
/******************************************************************************/

#include <iostream>
#include <vector>
#include <omp.h>
#include <cassert>

/**
@def ELEMENT_TYPE
@brief Tip elemenata niza
*/
#define ELEMENT_TYPE	double

/**
@def DEFAULT_SEED
@brief Podrazumevano pocetno seme za generator slucajnih brojava kojima se popunjava niz.
*/
#define DEFAULT_SEED	777

/**
@def DEFAULT_ARR_LEN
@brief Podrazumevana duzina niza
*/
#define DEFAULT_ARR_LEN	100

typedef std::vector<ELEMENT_TYPE> Array;

//**********************************************************

/**
@brief Glavna funkcija koja implementira algoritam.
*/
void quicksort_r(Array &A, int p, int r);

/**
@brief Pomocna funcija koja deli niz na 2 nova podniza i bira pivot.
*/
int  partition(Array &A, int p, int r);

/**
@brief Funkcija koja inicijalizuje niz pocetnim slucajnim vrednostima.
*/
void fillArray(Array &A);

/**
@brief Funckija koja proverava da li je niz sortiran.
*/
bool checkIfSorted(Array &A);
//**********************************************************


Array A;		    //Niz koji cemo sortirati
unsigned int arrLen = DEFAULT_ARR_LEN; //Duzina niza. Podrazumevana vrednost je 100.
int maxThreads;		    //Maksimalni dozvoljeni broj niti u programu.
int activeThreads = 1;	    //Broj aktivnih niti u programu.
int callCount = 0;	    //Broj pozivanja procedure Quicksort 
omp_lock_t threadCountLock; //Jednostavan lock; Koristi se prilikom ispitivanja broja aktivnih niti.

//**********************************************************

/*
@brief Command-line parametri: ime_programa [<max_broj_niti> [<broj_elemenata> <seed_za_rand>]]
Ukoliko se kompajlira bez podrske za OpenMP, broj niti ne igra nikakvu ulogu.
@param argc broj ulaznih argumenata
@param argv niz ulaznih parametara
@return 0 ako se program zavrsi ispravno.
*/
int main(int argc, char *argv[]){

    double start, stop;
	
	//Inicijalizacija generatora slucajnih brojeva
    std::srand(DEFAULT_SEED);
    
    //Maksimalni dozvoljeni broj niti je podrazumevano jednak broju dostupnih procesora
    maxThreads = omp_get_num_procs();


    //Ucitavamo zadate parametre sa komandne linije
    if (argc > 1) { 
	maxThreads = std::atoi(argv[1]);
	if (maxThreads<2 || maxThreads%2){
	    maxThreads = omp_get_num_procs();
	}
    }
    if (argc == 4) { 
	arrLen = std::atoi(argv[2]);
	std::srand(std::atoi(argv[3]));
    }

    //Na pocetku odmah rezervisemo kontinualni segment u memoriji za niz
    A.resize(arrLen);
    
    //Ispunimo niz elementima
    fillArray(A);
  
    //Podesavanje promenljivih okruzenja OpenMP-a; inicijalizacija lock-a
    omp_set_dynamic(false);
    omp_set_nested(true);
    omp_init_lock(&threadCountLock);

    //pocinjemo sa merenjem vremena
    start = omp_get_wtime();

    //Pozivamo Quicksort
    quicksort_r(A, 0, A.size()-1);
    
    //Kraj merenja
    stop = omp_get_wtime();

    //Provera da li je niz zaista sortiran
    if (!checkIfSorted(A)) {
	std::cout << "Greska! Niz nije sortiran!..." << std::endl;;
	return 1;
    }
    
    //Ispis podataka o izvrsavanju programa...
    std::cout << "Duzina niza: " << A.size() << std::endl;
    std::cout << "Broj pozivanja Quicksort: " << callCount << std::endl;
    std::cout << "Vreme izvrsavanja: " << (stop-start)*1000000 << "us = " << (stop-start)*1000 << "ms" << std::endl;

    return 0;
}// end main()

/* 
@brief Algoritam Quicksort. Rekurzivna verzija. Cormen et al. - "Introduction to Algorithms", 2nd Ed., poglavlje 7.1
@param A Niz koji treba sortirati
@param p Donja granica
@param r gornja granica
*/
void quicksort_r(Array &A, int p, int r){

    int q;
    
    #pragma omp atomic
    callCount++;

    if (p >= r) return;

    q = partition(A, p, r);
    
    //Ako je broj aktivnih niti vec jednak ili premasio dozvoljeni broj, nastavi sekvencijalno...
    #ifdef _OPENMP
	omp_set_lock(&threadCountLock);
    #endif
    //ovde zakljucavamo lock, dok 1 nit ne inkrementira threadCountLock.
    //Inace bi se moglo desiti da 2 niti ispune uslov, i da se kasnije premasi max. broj niti
    if (activeThreads >= maxThreads){
	#ifdef _OPENMP
	    omp_unset_lock(&threadCountLock);
	#endif
	quicksort_r(A, p, q-1);
	quicksort_r(A, q+1, r);
	return;
    }

	/* p,q i r stavljamo da su shared, ali to je samo da se ne prave uzalud privatne kopije, 
	   kad ce one samo trajati do poziva quicksort, kada ce se preneti po vrednosti.*/
	#pragma omp parallel num_threads(2) \
	                     shared(A, p, q, r, maxThreads, activeThreads, \
			            threadCountLock, callCount)
	{   
	    #pragma omp single nowait
	    {
		activeThreads++;
		#ifdef _OPENMP
		    omp_unset_lock(&threadCountLock);
		#endif
	    }
	    
	    assert(activeThreads <= maxThreads);

	    #pragma omp sections
	    {
		#pragma omp section
		    quicksort_r(A, p, q-1);
		#pragma omp section
		    quicksort_r(A, q+1, r);
	    }
	    #pragma omp single nowait
	    {
		#pragma omp atomic
		activeThreads--; //ovde je umesto lock-ovanja jeftinija operacija atomicno dekrementiranje
	    }
	    assert(activeThreads >= 1);
	}

}//end quicksort_r()

/* 
@brief Pomocni podprogram koji deli niz na dva nova i vraca pivot. 
@param A Niz koji treba sortirati
@param p Donja granica
@param r gornja granica
@return poziciju pivota
*/
int partition(Array &A, int p, int r){
    
    int i, j;
    ELEMENT_TYPE pivot;

    pivot = A[r];
    i = p - 1;

    for (j=p; j<r; j++){
	if (A[j] <= pivot) {
	    i++;
	    std::swap(A[i], A[j]);
	}
    }

    std::swap(A[i+1], A[r]);
    return i+1;
} //end partition()

/* 
@brief Funkcija koja inicijalizuje niz pocetnim slucajnim vrednostima. 
@param A Niz koji treba popuniti
*/
void fillArray(Array &A){
    Array::size_type i;
    for (i=0; i<arrLen; i++){
	A[i] = (ELEMENT_TYPE) std::rand();
	//std::cout << A[i] << " ";
    }
    //std::cout << std::endl;
}

/**
@brief Funckija koja proverava da li je niz sortiran.
@param A Niz
*/
bool checkIfSorted(Array &A) {
    Array::size_type i;
    ELEMENT_TYPE smaller = A[0];

    for (i=1; i<A.size(); i++){
	if (A[i]<smaller) return false;
	smaller = A[i];
    }

    return true;
}
/******************************************************************************/