/******************************************************************************/
/**
@file "pbs_openmp.c"
@brief Parallel bucket sort primer.

Primer sortiranja liste jednako distribuiranih brojeva koristeci OpenMP. 
Brojevi se nasumicno generisu izmedju 0 i 2*n. Koriste se Bucket sort i Sequential
sort algoritmi. Ovo je preradjena verzija istovetnog zadatka koji koristi MPI.
*/
/******************************************************************************/

#include <omp.h>
#include <stdio.h>
#include <stdlib.h>

/******************************************************************************/

void make_numbers(long int big_array[], int n, int n_bar, int p);

void sequential_sort(long int array[], int size);

int  get_min_pos(long int array[] , int eff_size);

/******************************************************************************/

/**
@brief Glavni program
@param argc Broj argumenata komandne linije
@param argv Argumenti komandne linije
@return Vraca @b 0 ako je sve uspesno izvrseno
*/
int main(int argc, char* argv[]) {
	
	long int* big_array;
   
	/* Podrazumevano se sortira 80 elemenata */
	int n = 80;
	int n_bar /* = n / p; */; /* broj elemenata koje pojedina nit dobija na obradu */
	int p; /* Broj niti na koje se deli obrada. Podrazumevano jednak broju dostupnih procesora */
	int i;	
	double start, stop;   /* Za merenje proteklog vremena */

	p = omp_get_num_procs();
		
	if (argc == 3) {
		n = atoi(argv[1]);
		p = atoi(argv[2]); 
	}
	
	/* Proverimo da li su parametri n i p validni */
	if (n % p != 0){
		fprintf(stderr,"The number of processes must evenly divide total elements.\n");
		exit(1);
	}
	
	n_bar = n / p;
        
	/* Napravimo veliki niz */
	big_array = malloc(n * sizeof(long int));
	
	if (big_array == NULL){
		fprintf(stderr, "Big_array malloc failed!!\n");
		exit(0);
	}

	printf("\nTotal elements = %d; Each process sorts %d elements.\n", 
		n, n_bar);

	/* Generisemo elemente niza */
	make_numbers(big_array, n, n_bar, p);
	
	/* Zapocinjemo merenje */
	start = omp_get_wtime();
	
	/* Svaka nit ce da sortira svoj deo niza, u zavisnosti od svog rednog broja */
	#pragma omp parallel num_threads(p) shared(big_array, n_bar)
	sequential_sort(&big_array[omp_get_thread_num()*n_bar], n_bar);
	
	/* Zaustavljamo merenje */
	stop = omp_get_wtime();


	printf("\nAfter sorting:\n");
	for(i = 0; i < n; i++)
		printf("%7ld %c", big_array[i], i % 8 == 7 ? '\n'  : ' '); 
	
	printf("\n\nTime to sort using %d threads = %lf msecs\n", p, (stop - start)/0.001);


	/* Oslobodimo memorijski prostor koji smo zauzeli */
	free(big_array);

	return 0;
}

/******************************************************************************/

/**
@brief Metoda koja generise elemente bafera.
@param big_array Bafer u koji se smestaju generisani elementi.
@param n Velicina bafera.
@param n_bar Koeficijent koji koristimo za generisanje elemenata (= n / p).
@param p Koeficijent koji koristimo za generisanje elemenata.
*/
void make_numbers(long int	big_array[],
		  int		n,
		  int		n_bar,
		  int		p)
{
	/* Stavljamo brojeve u "kofe", mada ih mozemo tretirati i drugacije */ 
	int i, q;
	
	printf("Before sorting:\n");
	for (q = 0; q < p; q++)
	{
		printf("\nT%d: ", q); 
		for (i = 0; i < n_bar; i++)
		{
			big_array[q * n_bar + i] =
				rand() % (2 * n / p) + (q * 2 * n / p);
			printf("%7ld %s",
			 big_array[q * n_bar + i], i % 8 == 7 ? "\n\t" : " "); 
		}
		printf("\n"); 
	}
	printf("\n");
}

/******************************************************************************/

/**
@brief Sequential sort algoritam.
@param array Bafer.
@param size Velicina bafera.
*/
void sequential_sort(long int array[], int size)
{
	/*
	Koristimo Sequential sort algoritam da sortiramo niz po rastucem
	poretku
	*/
	int eff_size, min_pos;
	long int temp;
	
	printf("Hello, I'm thread no. %d\n", omp_get_thread_num());

	for (eff_size = size; eff_size > 1; eff_size--)
	{
		min_pos = get_min_pos(array, eff_size);
		temp = array[min_pos];
		array[min_pos] = array[eff_size-1];
		array[eff_size-1] = temp;
	}
}

/******************************************************************************/

/**
@brief Vraca indeks najmanjeg od preostalih nesortiranih elemenata.
@param arr Bafer cije elemente treba sortirati.
@param eff_size Broj preostalih elemenata.
@return Vraca indeks najmanjeg od preostalih nesortiranih elemenata.
*/
int get_min_pos(long int arr[], int eff_size)
{
	int i, min_pos = 0;
	
	for (i=0; i < eff_size; i++)
		min_pos = arr[i] > arr[min_pos] ? i: min_pos;
	return min_pos;
}

/******************************************************************************/