/******************************************************************************/
/**
@file "Pthreads/C/Primer 4 - mutex/dot_product_threaded_mutex.c"
@brief Konkurentna verzija

Ovaj primer ilustruje upotrebu mutex promenljivih u programima sa nitima.
Do ove verzije se dolo modifikacijom sekvencijalne verzije programa 
(dot_product_simple.c), koja rauna skalarni proizvod dva vektora. Glavni
podaci su dostupni svim nitima preko globalno dostupne strukture. Glavna nit
eka da sve ostale niti zavre svoje proraune i onda na standardnom izlazu 
ispisuje rezultujuu sumu.@n @n

Autor ovog primera je Vijay Sonnad, IBM. Najnovija verzija koda, koju je kreirao
Blaise Barney, je preuzeta iz Pthreads tutorijala sa adrese
http://www.llnl.gov/computing/tutorials/pthreads/ i prilagodjen potrebama
ovog kursa.
*/
/******************************************************************************/

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

/******************************************************************************/

/**
@struct DOT_DATA
@brief Skup podataka nad kojima se vri obrada i rezultat izvravanja
*/
typedef struct {
    /**
    @property a
    @brief Pokaziva na prvi cinilac.
    */
	double* a;

    /**
    @property b
    @brief Pokaziva na drugi cinilac.
    */
	double* b;

    /**
    @property sum
    @brief Suma.
    */
	double sum; 

    /**
    @property vec_len
    @brief Duina inilaca.
    */
	int vec_len; 
} DOT_DATA;

/******************************************************************************/

/**
@def NUM_THRDS
@brief Koliko niti ce se koristiti.
*/
#define NUM_THRDS 4

/**
@def VEC_LEN
@brief Duina vektora.
*/
#define VEC_LEN 100

/**
@var dot_str
@brief Globalno dostupne neophodne informacije.
*/
DOT_DATA dot_str;

/**
@var call_thd
@brief Niz deskriptora niti.
*/
pthread_t call_thd[NUM_THRDS];

/**
@var mutex_sum
@brief Mutex promenljiva.
*/
pthread_mutex_t mutex_sum;

/******************************************************************************/

/**
@brief Posao koji svaka nit treba da obavlja.
@param arg Generiki pokaziva na broj niti.
@return void* Vraa rezultat izvravanja rutine ptrhead_exit().

Funkcija dot_prod se poziva kada se nit aktivira. Kao i ranije,
svi ulazni podaci bie dohvaeni pomou strukture tipa DOT_DATA
i svi dobijeni rezultati bie vraeni u istu strukturu.
Korist ovakvog pristupa je oigledna za vienitno programiranje:
Kada se nit kreira mi prosledjujemo jedan jedini argument pozvanoj funkciji -
tipino, to je broj niti. Svim ostalim potrebnim informacijama se pristupa
preko globalno dostupne strukture.
*/
void* dot_prod(void* arg)
{
	int i, start, end, offset, len ;
	int j = rand() * rand() * rand();
	double my_sum, *x, *y;
	offset = (int) arg;

	printf("Thread %x started\n", pthread_self());

	for (i = 0;  i < j; i++) i++;

	printf("Thread %x really started\n", pthread_self());

	len = dot_str.vec_len;
	start = offset * len;
	end = start + len;
	x = dot_str.a;
	y = dot_str.b;

	my_sum = 0;

	for (i = start; i < end ; i++) 
	{
		my_sum += (x[i] * y[i]);
	}

	// Pre azuriranja vrednosti polja deljene strukture zakljcamo
	// mutex, a onda ga kasnije ponovo otkljucamo po izlasku iz
	// kriticne sekcije.
	printf("Thread %x about to lock the mutex\n", pthread_self());
	pthread_mutex_lock (&mutex_sum);
	// Ulazimo u kriticnu sekciju.
	printf("Thread %x has entered critical section\n", pthread_self());

	for (i = 0;  i < j; i++) i++;

	dot_str.sum += my_sum;

	// Izlazimo iz kriticne sekcije.
	printf("Thread %x is now leaving critical section\n", pthread_self());
	pthread_mutex_unlock (&mutex_sum);
	printf("Thread %x unlocked the mutex\n", pthread_self());

	pthread_exit((void*) 0);
}

/******************************************************************************/

/**
@brief Glavni program.
@param argc Broj argumenata komandne linije.
@param argv Argumenti komandne linije.
@return int Rezultat izvravanja programa.

Glavni program kreira niti koje odrade ceo posao, a zatim, po zavretku,
ispie rezultat. Pre kreiranja niti stvaraju se ulazni podaci. Kako sve
niti auriraju deljenu strukturu, potrebna nam je mutex promenljiva za
medjusobno iskljuenje. Glavna nit eka da sve niti zavre posao, svaku od njih.
Eksplicitno specificiramo vrednost atributa niti koji dozvoljava glavnoj niti
da se pridrui nitima koje je kreirala. Takodje, treba obratiti panju na
oslobadjanje handler-a kada nam vie nisu potrebni.
*/
int main (int argc, char* argv[])
{
	int i;
	double* a, *b;
	int status;
	pthread_attr_t attr;

#ifdef WIN32
    pthread_win32_process_attach_np();
#endif

	a = (double*) malloc(NUM_THRDS * VEC_LEN * sizeof(double));
	b = (double*) malloc(NUM_THRDS * VEC_LEN * sizeof(double));

	for (i = 0; i < VEC_LEN * NUM_THRDS; i++)
	{
		a[i] = 1;
		b[i] = a[i];
	}

	dot_str.vec_len = VEC_LEN; 
	dot_str.a = a; 
	dot_str.b = b; 
	dot_str.sum = 0;

	pthread_mutex_init(&mutex_sum, NULL);

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	for (i = 0; i < NUM_THRDS; i++)
	{
		// Svaka nit funkcionise nad posebnim skupom podataka.
		// Offset je odredjen vrednoscu promenljive i.
		// Velicina podataka je odredjena vrednoscu
		// promenljive vec_len.
		pthread_create(&call_thd[i], &attr, dot_prod, (void*) i); 
	}

	pthread_attr_destroy(&attr);

	// Pridruzujemo glavnu nit ostalim nitima.
	for (i = 0; i < NUM_THRDS; i++)
	{
		pthread_join(call_thd[i], (void**) &status);
	}

	printf("Sum =  %f \n", dot_str.sum);

	free (a);
	free (b);
	pthread_mutex_destroy(&mutex_sum);
	pthread_exit(NULL);

#ifdef WIN32
    pthread_win32_process_detach_np();
#endif
}

/******************************************************************************/
